#include <stdexcept>
#include <iostream>

/*
 * This program implements a simple virtual machine.
 * The supported operations are listed in the enum OP.
 * The virtual machine is inspired by the IJVM https://en.wikipedia.org/wiki/IJVM
 * The given virtual machine is very simple. It can only execute integer operations.
 * Thanks to its simplicity, a program can be expressed as an array of integers,
 * as each operation and each constant can be expressed as an integer, cf. the
 * variable program.
 * When executing a program, see main() function, the virtual machine (applies to all VMs though)
 * follows the following pattern:
 *   - fetch next instruction
 *   - decode
 *   - execute
 * In our case, fetching the next instruction basically just means increasing the instruction
 * pointer ip to index the next instruction in the program array, cf. function fetch().
 * decode and execute both take place in the function eval(). First, via a switch statement,
 * we figure out the type of the given instruction, then we execute the instruction.
 */

int ip = 0; // instruction pointer. (a.k.a. program counter)
int sp = -1; // stack pointer
int stack [256];
bool running = true;


enum OP {
  ADD,
  SUB,
  MUL,
  DIV,
  DUPLICATETOP,
  LT,
  EQ,
  JMP,
  JMPIFTRUE,
  JMPIFFALSE,
  PRINTTOP,
  PUSH,
  POP,
  HALT,
  NOP // no operation
};


// the program to be executed goes here
// program as given executes:
//   for (int i = 111; i <= 122; ++i) {
//     std::cout << i << std::endl;
//   }
int program [] = {
  PUSH, 111,
  DUPLICATETOP, 
  PRINTTOP,  // JMPIFTRUE jumps here
  DUPLICATETOP,
  PUSH, 122,
  SUB,
  JMPIFFALSE, 16,
  PUSH, 1,
  ADD,
  JMP, 3,
  POP,  // JMP jumps here
  HALT
};

//int program [] = {
//  PUSH, 3,
//  PUSH, 4,
//  MUL,
//  PUSH, 5,
//  ADD,
//  PRINTTOP,
//  POP,
//  HALT
//};

// (5 + 3 * 4) / 2
//int program [] = {
//  PUSH, 2,
//  PUSH, 5,
//  PUSH, 3,
//  PUSH, 4,
//  MUL,
//  ADD,
//  DIV,
//  PRINTTOP,
//  POP,
//  HALT
//};

int A, B, C; // registers


void eval (int instruction) {
  switch (instruction) {
    case ADD:
      A = stack[sp];
      --sp;
      B = stack[sp];
      --sp;
      C = A + B;
      ++sp;
      stack[sp] = C;
      break;
    case SUB:
      A = stack[sp];
      --sp;
      B = stack[sp];
      --sp;
      C = A - B;
      ++sp;
      stack[sp] = C;
      break;
    case MUL:
      A = stack[sp];
      --sp;
      B = stack[sp];
      --sp;
      C = A * B;
      ++sp;
      stack[sp] = C;
      break;
    case DIV:
      A = stack[sp];
      --sp;
      B = stack[sp];
      --sp;
      C = A / B;
      ++sp;
      stack[sp] = C;
      break;
    case DUPLICATETOP:
      ++sp;
      stack[sp] = stack[sp-1];
      // std::cout << "DUPLICATETOP (sp=" << sp << ")" << std::endl;
      break;
    case LT:
      A = stack[sp];
      --sp;
      B = stack[sp];
      --sp;
      C = A < B; // 1 if true, 0 else
      ++sp;
      stack[sp] = C;
      break;
    case EQ:
      A = stack[sp];
      --sp;
      B = stack[sp];
      --sp;
      C = A == B; // 1 if true, 0 else
      ++sp;
      stack[sp] = C;
      break;
    case JMP:
      ++ip;
      ip = program[ip] - 1;    // -1 to compensate for next fetch
      break;
    case JMPIFTRUE:
      ++ip;
      if (stack[sp]) {
        ip = program[ip] - 1;  // -1 to compensate for next fetch
      }
      --sp;
      break;
    case JMPIFFALSE:
      ++ip;
      if (!stack[sp]) {
        ip = program[ip] - 1;  // -1 to compensate for next fetch
      }
      --sp;
      break;
    case PRINTTOP:
      std::cout << stack[sp] << std::endl;
      break;
    case PUSH:
      ++sp;
      stack[sp] = program[++ip];
      break;
    case POP:
      --sp;
      break;
    case HALT:
      running = false;
      //std::cout << "HALT: sp=" << sp << std::endl;
      break;
    case NOP:
      break;
    default:
      throw std::invalid_argument("Unknown Operation.");
  }
}


// fetch next instruction
int fetch () {
  return program[ip];
}


int main () {
  while (running) {
    auto op = fetch();
    eval(op);
    ++ip;
  }
}
